require 'attparser'


#===============================================================================
# CodeBlock
#
class CodeBlock
	# Version
	VERSION = '1.1'
	
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# First param should be the code that should be parsed. The remaining
	# options follow the names of the various properties that can be set. So,
	# for example, to create a block with named, starting and ending lines,
	# you could do this:
	#
	#    block = CodeBlock.new(src, 'start'=>'a', 'end'=>'b')
	#
	# You might find that you are often naming the starting line "start" and the
	# ending line "end". To simplify setting those common names, you can use the
	# se ("start/end") option:
	#    
	#    block = CodeBlock.new(src, 'se'=>true)
	#
	# That is exactly equivelant to:
	#
	#    block = CodeBlock.new(src, 'start'=>'start', 'end'=>'end')
	
	def initialize(raw, opts={})
		@lines_org = raw.lines
		@keep_meta = opts['keep_meta'] || false
		@tab = opts['tab'] || '    '
		@indent = opts['indent']
		@line_nums = opts['line_nums']
		@skip_skips = opts['skip_skips'] ? true : false
		
		# if 'se' param, set start and end to "start" and "end"
		if opts['se']
			@start = 'start'
			@end = 'end'
		else
			@start = opts['start']
			@end = opts['end']
		end
		
		# default collapse to true
		if opts.has_key?('collapse')
			@collapse = opts['collapse']
		else
			@collapse = true
		end
		
		# initialize block object
		@block_ob = nil
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# attribute readers
	#
	
	# The name of the line that starts the block. Leave nil to start at the
	# beginning of the file. If an ending line is given but that line is not
	# found then an exception is raised.
	attr_reader :start
	
	# The name of the line that ends the block. Leave nil for no ending line.
	# If an ending line is given but that line is not found then an exception is
	# raised.
	attr_reader :end
	
	# The string to convert tab characters to. Defaults to four spaces.
	attr_reader :tab
	
	# If set, indicates a string to put at the beginning of each line. For
	# Markdown, a value of four spaces is a good choice. Defaults to nil.
	attr_reader :indent
	
	# If true, then contiguous lines that contain just spaces are collapsed to a
	# single space-only line.
	attr_reader :collapse
	
	# If true, the line meta information (i.e. the ## and everything after it)
	# is not stripped from the output lines. Defaults to false.
	attr_reader :keep_meta
	
	
	# If true, then lines that are marked to be skipped are not skipped.
	# Defaults to false.
	attr_reader :skip_skips
	
	# If true, line numbers are output with the block of code. Defaults to
	# false.
	attr_reader :line_nums
	
	# The full array of lines in the code.
	attr_reader :lines_full
	
	#
	# attribute readers
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# start=, end=, tab=
	#
	
	# Sets the start property.
	def start=(new_val)
		@block_ob = nil
		@start = new_val
	end
	
	# Sets the end property.
	def end=(new_val)
		@block_ob = nil
		@end = new_val
	end
	
	# Sets the tab property.
	def tab=(new_val)
		@block_ob = nil
		@tab = new_val
	end
	
	# Sets the indent property.
	def indent=(new_val)
		@block_ob = nil
		@indent = new_val
	end
	
	# Sets the collapse property.
	def collapse=(new_val)
		@block_ob = nil
		@collapse = new_val
	end
	
	# Sets the keep_meta property.
	def keep_meta=(new_val)
		@block_ob = nil
		@keep_meta = new_val
	end
	
	# Sets the skip_skips property.
	def skip_skips=(new_val)
		@block_ob = nil
		@skip_skips = new_val
	end
	
	# Sets the line_nums property.
	def line_nums=(new_val)
		@block_ob = nil
		@line_nums = new_val
	end
	
	#
	# start=, end=, tab=
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# sort-of delegate some methods to block()
	# NOTE: I can't get delegate() to work for delegating to a method. Not sure
	# what I'm doing wrong. Feel free to fix it.
	#
	
	# Returns an array of the lines in the code block. Each element of the array
	# is a CodeBlock::Line object.
	def lines
		return block.lines
	end
	
	# Returns a string of the code block.
	def to_s
		return block.to_s
	end
	
	# Returns a string of the line numbers, one number per line.
	def line_nums_to_s
		return block.line_nums_to_s
	end
	
	# Returns a string of the notes for each each line. Notes are set with the
	# `notes` attribute in the meta section of the line, like this:
	#
	#    db['hero'] = 'Thor' ## notes="Set the value in the database"
	#
	#  There will be as many lines in the string as lines in the code block.
	#  Lines with no notes will be empty lines.
	def notes_to_s
		return block.notes_to_s
	end
	
	# A hash of named lines.
	def named
		return block.named
	end
	
	#
	# sort-of delegate some methods to block()
	#---------------------------------------------------------------------------
	
	
	# private
	private
	
	
	#---------------------------------------------------------------------------
	# block
	#
	def block()
		# build block if necessary
		if not @block_ob
			@lines_full = @lines_org.map {|line| CodeBlock::Line.new(self, line)}
			opts = {}
			opts['start'] = @start
			opts['end'] = @end
			opts['tab'] = @tab
			opts['collapse'] = @collapse
			@block_ob = CodeBlock::Block.new(self, opts)
		end
		
		# return
		return @block_ob
	end
	#
	# block
	#---------------------------------------------------------------------------
end
#
# CodeBlock
#===============================================================================


#===============================================================================
# CodeBlock::Line
#

# A CodeBlock::Line represents a single line in the block of code. Generally you
# will not need to instantiate a CodeBlock::Line object yourself.
class CodeBlock::Line
	#---------------------------------------------------------------------------
	# initialize
	#
	
	# Accepts a CodeBlock object and the raw string from the code.
	def initialize(code, raw)
		@code = code
		@raw = raw
		@number = nil
		@display = @raw
		
		# unless keep meta, else strip ## and everything after
		if @code.keep_meta
			@display = @raw
		else
			@display = @raw.sub(/\s*\#\#.*/mu, '')
		end
		
		# remove trailing spaces
		@display = @display.rstrip
		
		# attributes
		if @raw.match(/\#\#/mu)
			raw_atts = @raw.sub(/\A.*\#\#\s*/mu, '')
			raw_atts = raw_atts.rstrip
			
			# if raw atts are a single string without =, use that as the name
			if raw_atts.match(/\A\S+\z/mu) and (not raw_atts.match(/\=/mu))
				@atts = {'name'=>raw_atts}
			else
				@atts = AttParser.parse(raw_atts, 'infer'=>true)
			end
		else
			@atts = {}
		end
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# readers and accessors
	#
	
	# Returns a hash of the name=value pairs set in the line's meta section.
	attr_reader :atts
	
	# Returns the line's number in the code block.
	attr_accessor :number
	
	#
	# readers and accessors
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# convenience
	#
	
	# Returns the value of 'name' in the line's attributes. Returns nil if no
	# name was set.
	def name
		return @atts['name']
	end
	
	# Returns the value of 'skip' in the line's attributes. Returns nil if no
	# skip was set.
	def skip
		return @atts['skip']
	end
	
	# Returns the value of 'notes' in the line's attributes. Returns nil if no
	# notes were set.
	def notes
		return @atts['notes']
	end
	
	#
	# convenience
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# to_s
	#
	
	# Returns the line as it should be displayed, taking into account tabs,
	# indentation, line number, etc. If the line is to be skipped then nil is
	# returned.
	def to_s
		if @atts['skip'] and (not @code.skip_skips)
			return nil
		else
			return @display
		end
	end
	#
	# to_s
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# has_content?
	#
	
	# Returns true if the line has any non-space characters.
	def has_content?
		return @display.match(/\S/mu)
	end
	
	# Returns false if the line has any non-space characters.
	def no_content?
		return !has_content?
	end
	
	#
	# has_content?
	#---------------------------------------------------------------------------
end
#
# CodeBlock::Line
#===============================================================================



#===============================================================================
# CodeBlock::Block
#

# This is the class that does the work of filtering which lines should be
# displayed and which shouldn't. Generally you won't have to work directly with
# this class.
class CodeBlock::Block
	#---------------------------------------------------------------------------
	# initialize
	#
	def initialize(p_code, opts={})
		@code = p_code
		@collapse = opts['collapse']
		@tab = opts['tab'] || '    '
		@start = opts['start']
		@end = opts['end']
		
		# build lines array
		build_lines()
	end
	#
	# initialize
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# attribute readers
	#
	
	# The array of CodeBlock::Line objects.
	attr_reader :lines
	
	# A hash of named lines.
	attr_reader :named
	
	#
	# attribute readers
	#---------------------------------------------------------------------------
	
	
	
	#---------------------------------------------------------------------------
	# to_s
	#
	
	# Builds the string that is returned by CodeBlock#to_s.
	def to_s()
		# $tm.hrm
		rv = []
		indent = @code.indent || ''
		line_nums = @code.line_nums
		
		# get maximum width
		if line_nums
			max_width = @lines[-1].number.to_s.length
		end
		
		# build return array
		@lines.each do |line|
			# get string
			v = line.to_s
			
			# substitute tabs
			if @tab
				v = v.gsub(/\t/mu, @tab)
			end
			
			# line numbers
			if line_nums
				v = line.number.to_s.rjust(max_width) + '.  ' + v
			end
			
			# indent
			v = indent + v
			
			# add to return array
			rv.push v
		end
		
		# return
		return rv.join("\n")
	end
	#
	# to_s
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# line_nums_to_s
	#
	
	# Builds the string that is returned by CodeBlock#line_nums_to_s.
	def line_nums_to_s
		# $tm.hrm
		rv = []
		
		# get maximum width
		width = @lines[-1].number.to_s.length
		
		# loop through lines
		@lines.each do |line|
			rv.push line.number.to_s.rjust(width) + '.'
		end
		
		# return
		return rv.join("\n")
	end
	#
	# line_nums_to_s
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# notes_to_s
	#
	
	# Builds the string that is returned by CodeBlock#notes_to_s.
	def notes_to_s
		rv = []
		
		# loop through lines
		@lines.each do |line|
			rv.push line.notes || ''
		end
		
		# return
		return rv.join("\n")
	end
	#
	# notes_to_s
	#---------------------------------------------------------------------------
	
	
	# private
	private
	
	
	#---------------------------------------------------------------------------
	# build_lines
	#
	def build_lines
		# $tm.hrm
		@lines = []
		
		# starting line name
		started = @start ? false : true
		ended = false
		
		# get visible lines
		@code.lines_full.each do |line|
			# start if necessary
			if (not started) and @start
				started = line.atts['name'] == @start
			end
			
			# add line
			if started and (not ended)
				if line.to_s
					@lines.push line.clone
				end
			end
			
			# if ending line
			if (not ended) and @end
				ended = line.atts['name'] == @end
			end
		end
		
		# strip leading empty lines
		while @lines.any? and (not @lines[0].to_s.match(/\S/mu))
			@lines.shift
		end
		
		# strip trailing empty lines
		while @lines.any? and (not @lines[-1].to_s.match(/\S/mu))
			@lines.pop
		end
		
		# collapse
		if @collapse
			while collapse_find()
			end
		end
		
		# number lines
		@lines.each_with_index() do |line, idx|
			line.number = idx + 1
		end
		
		# build hash of named lines
		@named = {}
		
		# get named lines
		@lines.each do |line|
			if line.name
				if @named[line.name]
					raise 'https://motley.uno/messages/code/init/redundant-name: ' + line.name
				end
				
				@named[line.name] = line
			end
		end
	end
	#
	# build_lines
	#---------------------------------------------------------------------------
	
	
	#---------------------------------------------------------------------------
	# collapse_find
	#
	def collapse_find()
		# loop through lines
		@lines.each_with_index do|line, idx|
			if idx > 0
				if line.no_content? and @lines[idx-1].no_content?
					@lines.delete_at(idx-1)
					return true
				end
			end
		end
		
		# didn't find anymore lines to collapse
		return false
	end
	#
	# collapse_find
	#---------------------------------------------------------------------------
end
#
# CodeBlock::Block
#===============================================================================